/* eslint-disable no-param-reassign,react/display-name */
import React, { createContext, memo, useCallback, useContext, useEffect, useLayoutEffect, useMemo, useRef, } from 'react';
import ReactDOM from 'react-dom';
import { useSyncExternalStore } from 'use-sync-external-store/shim';
import { shallow } from './shallow';
const batchedUpdates = ReactDOM.unstable_batchedUpdates || ((fn) => fn());
const useIsomorphicLayoutEffect = typeof window === 'undefined' ? useEffect : useLayoutEffect;
/**
 * Create a store with `useHook`
 */
export function createStore(useHook) {
    // Keep the Context never triggering an update
    const Context = createContext(null);
    const Provider = memo(({ initialState, children }) => {
        const value = useHook(initialState);
        const bridge = useRef({ value, listeners: new Set() }).current;
        bridge.value = value;
        useIsomorphicLayoutEffect(() => batchedUpdates(() => bridge.listeners.forEach((listener) => listener(value))));
        return (React.createElement(Context.Provider, { value: bridge }, typeof children === 'function' ? children(value) : children));
    });
    function useContextValue() {
        const value = useContext(Context);
        if (value === null) {
            throw new Error('The component must be wrapped by <Store.Provider />');
        }
        return value;
    }
    function useStoreUpdate(bridge, equalityFn) {
        const fnRef = useRef(equalityFn);
        fnRef.current = equalityFn;
        const subscribe = useCallback((update) => {
            const { listeners } = bridge;
            function listener(nextValue) {
                if (!fnRef.current(nextValue)) {
                    update();
                }
            }
            listeners.add(listener);
            return () => listeners.delete(listener);
        }, []);
        const getSnapshot = useCallback(() => bridge.value, []);
        useSyncExternalStore(subscribe, getSnapshot);
    }
    function useSelector(selector) {
        const bridge = useContextValue();
        const selected = selector(bridge.value);
        useStoreUpdate(bridge, (nextValue) => shallow(selected, selector(nextValue)));
        return selected;
    }
    function usePicker() {
        const bridge = useContextValue();
        const proxy = useRef();
        const canUpdate = useRef(false);
        const updated = useRef(false);
        useMemo(() => {
            const handler = {
                get(target, key) {
                    canUpdate.current = true;
                    target[key] = bridge.value[key];
                    return target[key];
                },
                set(target, key, val) {
                    if (key in target && val !== target[key]) {
                        updated.current = true;
                        target[key] = val;
                    }
                    return true;
                },
            };
            proxy.current = new Proxy({}, handler);
        }, []);
        useStoreUpdate(bridge, (nextValue) => {
            if (!canUpdate.current)
                return true;
            // Trigger update
            Object.assign(proxy.current, nextValue);
            const result = !updated.current;
            updated.current = false;
            return result;
        });
        return proxy.current;
    }
    function withProvider(Children, initialState) {
        return function ProviderComponent(props) {
            return (React.createElement(Provider, { initialState: initialState },
                React.createElement(Children, { ...props })));
        };
    }
    return {
        Context,
        Provider,
        useSelector,
        usePicker,
        withProvider,
    };
}
